@@ -6,20 +6,48 @@ use crate::{
     parser::Rule,
 };
 
-use sway_types::span::Span;
+use sway_types::{
+    span::{join_spans, Span},
+    Ident,
+};
 
 use pest::iterators::Pair;
 
+/// Represents the left hand side of a reassignment, which could either be a regular variable
+/// expression, denoted by [ReassignmentTarget::VariableExpression], or, a storage field, denoted
+/// by [ReassignmentTarget::StorageField].
+#[derive(Debug, Clone)]
+pub enum ReassignmentTarget {
+    VariableExpression(Box<Expression>),
+    StorageField(Vec<Ident>),
+}
+
 #[derive(Debug, Clone)]
 pub struct Reassignment {
     // the thing being reassigned
-    pub lhs: Box<Expression>,
+    pub lhs: ReassignmentTarget,
     // the expression that is being assigned to the lhs
     pub rhs: Expression,
     pub(crate) span: Span,
 }
 
 impl Reassignment {
+    pub fn lhs_span(&self) -> Span {
+        match &self.lhs {
+            ReassignmentTarget::VariableExpression(var) => match **var {
+                Expression::SubfieldExpression { ref span, .. } => span.clone(),
+                Expression::VariableExpression { ref name, .. } => name.span().clone(),
+                _ => {
+                    unreachable!("any other reassignment lhs is invalid and cannot be constructed.")
+                }
+            },
+            ReassignmentTarget::StorageField(ref idents) => {
+                idents.iter().fold(idents[0].span().clone(), |acc, ident| {
+                    join_spans(acc, ident.span().clone())
+                })
+            }
+        }
+    }
     pub(crate) fn parse_from_pair(
         pair: Pair<Rule>,
         config: Option<&BuildConfig>,
@@ -35,31 +63,25 @@ impl Reassignment {
         let variable_or_struct_reassignment = iter.next().expect("guaranteed by grammar");
         match variable_or_struct_reassignment.as_rule() {
             Rule::variable_reassignment => {
-                let mut iter = variable_or_struct_reassignment.into_inner();
-                let name_result = check!(
-                    Expression::parse_from_pair_inner(iter.next().unwrap(), config),
+                let (name_result, mut body_result) = check!(
+                    parse_simple_variable_reassignment(
+                        variable_or_struct_reassignment,
+                        config,
+                        path,
+                    ),
                     return err(warnings, errors),
                     warnings,
                     errors
                 );
-                let body = iter.next().unwrap();
-                let mut body_result = check!(
-                    Expression::parse_from_pair(body.clone(), config),
-                    ParserLifter::empty(error_recovery_exp(Span {
-                        span: body.as_span(),
-                        path
-                    })),
-                    warnings,
-                    errors
-                );
 
                 let mut var_decls = name_result.var_decls;
                 var_decls.append(&mut body_result.var_decls);
                 let reassign = Reassignment {
-                    lhs: Box::new(name_result.value),
+                    lhs: ReassignmentTarget::VariableExpression(Box::new(name_result.value)),
                     rhs: body_result.value,
                     span,
                 };
+
                 ok(
                     ParserLifter {
                         var_decls,
@@ -70,67 +92,21 @@ impl Reassignment {
                 )
             }
             Rule::struct_field_reassignment => {
-                let mut iter = variable_or_struct_reassignment.into_inner();
-                let lhs = iter.next().expect("guaranteed by grammar");
-                let rhs = iter.next().expect("guaranteed by grammar");
-                let rhs_span = Span {
-                    span: rhs.as_span(),
-                    path: path.clone(),
-                };
-                let body_result = check!(
-                    Expression::parse_from_pair(rhs, config),
-                    ParserLifter::empty(error_recovery_exp(rhs_span)),
-                    warnings,
-                    errors
-                );
-
-                let inner = lhs.into_inner().next().expect("guaranteed by grammar");
-                assert_eq!(inner.as_rule(), Rule::subfield_path);
-
-                // treat parent as one expr, final name as the field to be accessed
-                // if there are multiple fields, this is a nested expression
-                // i.e. `a.b.c` is a lookup of field `c` on `a.b` which is a lookup
-                // of field `b` on `a`
-                // the first thing is either an exp or a var, everything subsequent must be
-                // a field
-                let mut name_parts = inner.into_inner();
-                let mut expr_result = check!(
-                    parse_subfield_path_ensure_only_var(
-                        name_parts.next().expect("guaranteed by grammar"),
-                        config
-                    ),
+                let (mut expr_result, body_result) = check!(
+                    parse_subfield_reassignment(variable_or_struct_reassignment, config, path),
                     return err(warnings, errors),
                     warnings,
                     errors
                 );
 
-                for name_part in name_parts {
-                    let expr = Expression::SubfieldExpression {
-                        prefix: Box::new(expr_result.value.clone()),
-                        span: Span {
-                            span: name_part.as_span(),
-                            path: path.clone(),
-                        },
-                        field_to_access: check!(
-                            ident::parse_from_pair(name_part, config),
-                            continue,
-                            warnings,
-                            errors
-                        ),
-                    };
-                    expr_result = ParserLifter {
-                        var_decls: expr_result.var_decls,
-                        value: expr,
-                    };
-                }
-
                 let mut var_decls = body_result.var_decls;
                 var_decls.append(&mut expr_result.var_decls);
                 let exp = Reassignment {
-                    lhs: Box::new(expr_result.value),
+                    lhs: ReassignmentTarget::VariableExpression(Box::new(expr_result.value)),
                     rhs: body_result.value,
                     span,
                 };
+
                 ok(
                     ParserLifter {
                         var_decls,
@@ -140,11 +116,142 @@ impl Reassignment {
                     errors,
                 )
             }
+            Rule::storage_reassignment => {
+                // there is some ugly handling of the pest data structure here
+                // because we are reusing code from struct field reassignments
+                // this leads to some nested `Rule`s, so the actual storage
+                // reassignment requires some `expect()`s to extract.
+                let mut parts = variable_or_struct_reassignment.into_inner();
+                let _storage_keyword = parts.next();
+                let mut parts = parts.collect::<Vec<_>>();
+                let rhs = parts.pop();
+                assert_eq!(rhs.as_ref().map(|x| x.as_rule()), Some(Rule::expr));
+                let rhs = rhs.expect("guaranteed by grammar");
+                let rhs = check!(
+                    Expression::parse_from_pair(rhs, config),
+                    return err(warnings, errors),
+                    warnings,
+                    errors
+                );
+                let mut lhs = Vec::new();
+                for item in parts {
+                    lhs.push(check!(
+                        ident::parse_from_pair(item, config),
+                        continue,
+                        warnings,
+                        errors
+                    ))
+                }
+
+                let reassign = Reassignment {
+                    lhs: ReassignmentTarget::StorageField(lhs),
+                    rhs: rhs.value,
+                    span,
+                };
+                ok(
+                    ParserLifter {
+                        var_decls: rhs.var_decls,
+                        value: reassign,
+                    },
+                    warnings,
+                    errors,
+                )
+            }
             _ => unreachable!("guaranteed by grammar"),
         }
     }
 }
 
+fn parse_simple_variable_reassignment(
+    pair: Pair<Rule>,
+    config: Option<&BuildConfig>,
+    path: Option<std::sync::Arc<std::path::PathBuf>>,
+) -> CompileResult<(ParserLifter<Expression>, ParserLifter<Expression>)> {
+    let mut warnings = vec![];
+    let mut errors = vec![];
+    let mut iter = pair.into_inner();
+    let name_result = check!(
+        Expression::parse_from_pair(iter.next().unwrap(), config),
+        return err(warnings, errors),
+        warnings,
+        errors
+    );
+    let body = iter.next().unwrap();
+    let body_result = check!(
+        Expression::parse_from_pair(body.clone(), config),
+        ParserLifter::empty(error_recovery_exp(Span {
+            span: body.as_span(),
+            path
+        })),
+        warnings,
+        errors
+    );
+
+    ok((name_result, body_result), warnings, errors)
+}
+
+fn parse_subfield_reassignment(
+    pair: Pair<Rule>,
+    config: Option<&BuildConfig>,
+    path: Option<std::sync::Arc<std::path::PathBuf>>,
+) -> CompileResult<(ParserLifter<Expression>, ParserLifter<Expression>)> {
+    let mut warnings = vec![];
+    let mut errors = vec![];
+    let mut iter = pair.into_inner();
+    let lhs = iter.next().expect("guaranteed by grammar");
+    let rhs = iter.next().expect("guaranteed by grammar");
+    let rhs_span = Span {
+        span: rhs.as_span(),
+        path: path.clone(),
+    };
+    let body_result = check!(
+        Expression::parse_from_pair(rhs, config),
+        ParserLifter::empty(error_recovery_exp(rhs_span)),
+        warnings,
+        errors
+    );
+    let inner = lhs.into_inner().next().expect("guaranteed by grammar");
+    assert_eq!(inner.as_rule(), Rule::subfield_path);
+    // treat parent as one expr, final name as the field to be accessed
+    // if there are multiple fields, this is a nested expression
+    // i.e. `a.b.c` is a lookup of field `c` on `a.b` which is a lookup
+    // of field `b` on `a`
+    // the first thing is either an exp or a var, everything subsequent must be
+    // a field
+    let mut name_parts = inner.into_inner();
+    let mut expr_result = check!(
+        parse_subfield_path_ensure_only_var(
+            name_parts.next().expect("guaranteed by grammar"),
+            config
+        ),
+        return err(warnings, errors),
+        warnings,
+        errors
+    );
+
+    for name_part in name_parts {
+        let expr = Expression::SubfieldExpression {
+            prefix: Box::new(expr_result.value.clone()),
+            span: Span {
+                span: name_part.as_span(),
+                path: path.clone(),
+            },
+            field_to_access: check!(
+                ident::parse_from_pair(name_part, config),
+                continue,
+                warnings,
+                errors
+            ),
+        };
+        expr_result = ParserLifter {
+            var_decls: expr_result.var_decls,
+            value: expr,
+        };
+    }
+
+    ok((expr_result, body_result), warnings, errors)
+}
+
 fn parse_subfield_path_ensure_only_var(
     item: Pair<Rule>,
     config: Option<&BuildConfig>,
