@@ -4,18 +4,19 @@ use crate::{
     build_config::BuildConfig,
     parse_tree::OwnedCallPath,
     semantic_analysis::ast_node::{TypedEnumVariant, TypedStructField},
-    Ident, Rule, TypeParameter,
+    Ident, Rule, TypeArgument, TypeParameter,
 };
 
 use sway_types::span::Span;
 
 use derivative::Derivative;
-
 use pest::iterators::Pair;
+use std::hash::{Hash, Hasher};
+
 /// Type information without an associated value, used for type inferencing and definition.
 // TODO use idents instead of Strings when we have arena spans
 #[derive(Derivative)]
-#[derivative(Debug, Clone, Eq, PartialEq, Hash)]
+#[derivative(Debug, Clone)]
 pub enum TypeInfo {
     Unknown,
     UnknownGeneric {
@@ -35,7 +36,7 @@ pub enum TypeInfo {
     /// For the type inference engine to use when a type references another type
     Ref(TypeId),
 
-    Tuple(Vec<TypeId>),
+    Tuple(Vec<TypeArgument>),
     /// Represents a type which contains methods to issue a contract call.
     /// The specific contract is identified via the `Ident` within.
     ContractCaller {
@@ -52,6 +53,7 @@ pub enum TypeInfo {
     /// until the semantic analysis stage.
     Custom {
         name: Ident,
+        type_arguments: Vec<TypeArgument>,
     },
     SelfType,
     Byte,
@@ -64,8 +66,183 @@ pub enum TypeInfo {
     ErrorRecovery,
     // Static, constant size arrays.
     Array(TypeId, usize),
+    /// Represents the entire storage declaration struct
+    /// Stored without initializers here, as typed struct fields,
+    /// so type checking is able to treat it as a struct with fields.
+    Storage {
+        fields: Vec<TypedStructField>,
+    },
 }
 
+// NOTE: Hash and PartialEq must uphold the invariant:
+// k1 == k2 -> hash(k1) == hash(k2)
+// https://doc.rust-lang.org/std/collections/struct.HashMap.html
+impl Hash for TypeInfo {
+    fn hash<H: Hasher>(&self, state: &mut H) {
+        match self {
+            TypeInfo::Str(len) => {
+                state.write_u8(1);
+                len.hash(state);
+            }
+            TypeInfo::UnsignedInteger(bits) => {
+                state.write_u8(2);
+                bits.hash(state);
+            }
+            TypeInfo::Numeric => {
+                state.write_u8(3);
+            }
+            TypeInfo::Boolean => {
+                state.write_u8(4);
+            }
+            TypeInfo::Tuple(fields) => {
+                state.write_u8(5);
+                fields.hash(state);
+            }
+            TypeInfo::Byte => {
+                state.write_u8(6);
+            }
+            TypeInfo::B256 => {
+                state.write_u8(7);
+            }
+            TypeInfo::Enum {
+                name,
+                variant_types,
+            } => {
+                state.write_u8(8);
+                name.hash(state);
+                variant_types.hash(state);
+            }
+            TypeInfo::Struct { name, fields } => {
+                state.write_u8(9);
+                name.hash(state);
+                fields.hash(state);
+            }
+            TypeInfo::ContractCaller { abi_name, address } => {
+                state.write_u8(10);
+                abi_name.hash(state);
+                address.hash(state);
+            }
+            TypeInfo::Contract => {
+                state.write_u8(11);
+            }
+            TypeInfo::ErrorRecovery => {
+                state.write_u8(12);
+            }
+            TypeInfo::Unknown => {
+                state.write_u8(13);
+            }
+            TypeInfo::SelfType => {
+                state.write_u8(14);
+            }
+            TypeInfo::UnknownGeneric { name } => {
+                state.write_u8(15);
+                name.hash(state);
+            }
+            TypeInfo::Custom {
+                name,
+                type_arguments,
+            } => {
+                state.write_u8(16);
+                name.hash(state);
+                type_arguments.hash(state);
+            }
+            TypeInfo::Ref(id) => {
+                state.write_u8(17);
+                look_up_type_id(*id).hash(state);
+            }
+            TypeInfo::Array(elem_ty, count) => {
+                state.write_u8(18);
+                look_up_type_id(*elem_ty).hash(state);
+                count.hash(state);
+            }
+            TypeInfo::Storage { fields } => {
+                state.write_u8(19);
+                fields.hash(state);
+            }
+        }
+    }
+}
+
+// NOTE: Hash and PartialEq must uphold the invariant:
+// k1 == k2 -> hash(k1) == hash(k2)
+// https://doc.rust-lang.org/std/collections/struct.HashMap.html
+impl PartialEq for TypeInfo {
+    fn eq(&self, other: &Self) -> bool {
+        match (self, other) {
+            (Self::Unknown, Self::Unknown) => true,
+            (Self::Boolean, Self::Boolean) => true,
+            (Self::SelfType, Self::SelfType) => true,
+            (Self::Byte, Self::Byte) => true,
+            (Self::B256, Self::B256) => true,
+            (Self::Numeric, Self::Numeric) => true,
+            (Self::Contract, Self::Contract) => true,
+            (Self::ErrorRecovery, Self::ErrorRecovery) => true,
+            (Self::UnknownGeneric { name: l }, Self::UnknownGeneric { name: r }) => l == r,
+            (
+                Self::Custom {
+                    name: l_name,
+                    type_arguments: l_type_args,
+                },
+                Self::Custom {
+                    name: r_name,
+                    type_arguments: r_type_args,
+                },
+            ) => l_name == r_name && l_type_args == r_type_args,
+            (Self::Str(l), Self::Str(r)) => l == r,
+            (Self::UnsignedInteger(l), Self::UnsignedInteger(r)) => l == r,
+            (
+                Self::Enum {
+                    name: l_name,
+                    variant_types: l_variant_types,
+                    ..
+                },
+                Self::Enum {
+                    name: r_name,
+                    variant_types: r_variant_types,
+                    ..
+                },
+            ) => l_name == r_name && l_variant_types == r_variant_types,
+            (
+                Self::Struct {
+                    name: l_name,
+                    fields: l_fields,
+                    ..
+                },
+                Self::Struct {
+                    name: r_name,
+                    fields: r_fields,
+                    ..
+                },
+            ) => l_name == r_name && l_fields == r_fields,
+            (Self::Ref(l), Self::Ref(r)) => look_up_type_id(*l) == look_up_type_id(*r),
+            (Self::Tuple(l), Self::Tuple(r)) => l
+                .iter()
+                .zip(r.iter())
+                .map(|(l, r)| look_up_type_id(l.type_id) == look_up_type_id(r.type_id))
+                .all(|x| x),
+            (
+                Self::ContractCaller {
+                    abi_name: l_abi_name,
+                    address: l_address,
+                },
+                Self::ContractCaller {
+                    abi_name: r_abi_name,
+                    address: r_address,
+                },
+            ) => l_abi_name == r_abi_name && l_address == r_address,
+            (Self::Array(l0, l1), Self::Array(r0, r1)) => {
+                look_up_type_id(*l0) == look_up_type_id(*r0) && l1 == r1
+            }
+            (TypeInfo::Storage { fields: l_fields }, TypeInfo::Storage { fields: r_fields }) => {
+                l_fields == r_fields
+            }
+            _ => false,
+        }
+    }
+}
+
+impl Eq for TypeInfo {}
+
 impl Default for TypeInfo {
     fn default() -> Self {
         TypeInfo::Unknown
@@ -74,60 +251,48 @@ impl Default for TypeInfo {
 
 impl TypeInfo {
     pub(crate) fn parse_from_pair(
-        input: Pair<Rule>,
-        config: Option<&BuildConfig>,
-    ) -> CompileResult<Self> {
-        match input.as_rule() {
-            Rule::type_name | Rule::generic_type_param => (),
-            _ => {
-                let span = Span {
-                    span: input.as_span(),
-                    path: config.map(|config| config.dir_of_code.clone()),
-                };
-                let errors = vec![CompileError::Internal(
-                    "Unexpected token while parsing type.",
-                    span,
-                )];
-                return err(vec![], errors);
-            }
-        }
-        Self::parse_from_pair_inner(input.into_inner().next().unwrap(), config)
-    }
-
-    fn parse_from_pair_inner(
-        input: Pair<Rule>,
+        type_name_pair: Pair<Rule>,
         config: Option<&BuildConfig>,
     ) -> CompileResult<Self> {
         let mut warnings = vec![];
         let mut errors = vec![];
+        let mut iter = type_name_pair.into_inner();
+        let input = iter.next().unwrap();
         let span = Span {
             span: input.as_span(),
             path: config.map(|config| config.dir_of_code.clone()),
         };
         let type_info = match input.as_rule() {
             Rule::str_type => {
-                check!(
+                let type_info = check!(
                     parse_str_type(input.as_str(), span),
                     return err(warnings, errors),
                     warnings,
                     errors
-                )
+                );
+                type_info
+            }
+            Rule::ident => {
+                let type_info = TypeInfo::pair_as_str_to_type_info(input, config);
+                match iter.next() {
+                    Some(types) => match type_info {
+                        TypeInfo::Custom { name, .. } => {
+                            let type_arguments = check!(
+                                TypeArgument::parse_arguments_from_pair(types, config),
+                                vec!(),
+                                warnings,
+                                errors
+                            );
+                            TypeInfo::Custom {
+                                name,
+                                type_arguments,
+                            }
+                        }
+                        _ => unimplemented!(),
+                    },
+                    None => type_info,
+                }
             }
-            Rule::ident | Rule::generic_type_param => match input.as_str().trim() {
-                "u8" => TypeInfo::UnsignedInteger(IntegerBits::Eight),
-                "u16" => TypeInfo::UnsignedInteger(IntegerBits::Sixteen),
-                "u32" => TypeInfo::UnsignedInteger(IntegerBits::ThirtyTwo),
-                "u64" => TypeInfo::UnsignedInteger(IntegerBits::SixtyFour),
-                "bool" => TypeInfo::Boolean,
-                "unit" => TypeInfo::Tuple(Vec::new()),
-                "byte" => TypeInfo::Byte,
-                "b256" => TypeInfo::B256,
-                "Self" | "self" => TypeInfo::SelfType,
-                "Contract" => TypeInfo::Contract,
-                _other => TypeInfo::Custom {
-                    name: Ident::new(span),
-                },
-            },
             Rule::array_type => {
                 let mut array_inner_iter = input.into_inner();
                 let elem_type_info = match array_inner_iter.next() {
@@ -141,7 +306,7 @@ impl TypeInfo {
                     Some(array_elem_type_pair) => {
                         check!(
                             Self::parse_from_pair(array_elem_type_pair, config),
-                            return err(warnings, errors),
+                            TypeInfo::ErrorRecovery,
                             warnings,
                             errors
                         )
@@ -198,18 +363,13 @@ impl TypeInfo {
                 TypeInfo::Array(insert_type(elem_type_info), elem_count)
             }
             Rule::tuple_type => {
-                let mut field_type_ids = vec![];
-                for field in input.into_inner() {
-                    let field_type = check!(
-                        TypeInfo::parse_from_pair(field, config),
-                        TypeInfo::Tuple(Vec::new()),
-                        warnings,
-                        errors
-                    );
-                    let field_type_id = crate::type_engine::insert_type(field_type);
-                    field_type_ids.push(field_type_id);
-                }
-                TypeInfo::Tuple(field_type_ids)
+                let fields = check!(
+                    TypeArgument::parse_arguments_from_pair(input, config),
+                    return err(warnings, errors),
+                    warnings,
+                    errors
+                );
+                TypeInfo::Tuple(fields)
             }
             _ => {
                 errors.push(CompileError::Internal(
@@ -222,6 +382,52 @@ impl TypeInfo {
         ok(type_info, warnings, errors)
     }
 
+    pub(crate) fn parse_from_type_param_pair(
+        input: Pair<Rule>,
+        config: Option<&BuildConfig>,
+    ) -> CompileResult<Self> {
+        let warnings = vec![];
+        let mut errors = vec![];
+        let type_info = match input.as_rule() {
+            Rule::type_param => Self::pair_as_str_to_type_info(input, config),
+            _ => {
+                let span = Span {
+                    span: input.as_span(),
+                    path: config.map(|config| config.dir_of_code.clone()),
+                };
+                errors.push(CompileError::Internal(
+                    "Unexpected token while parsing type.",
+                    span,
+                ));
+                return err(warnings, errors);
+            }
+        };
+        ok(type_info, warnings, errors)
+    }
+
+    pub fn pair_as_str_to_type_info(input: Pair<Rule>, config: Option<&BuildConfig>) -> Self {
+        let span = Span {
+            span: input.as_span(),
+            path: config.map(|config| config.dir_of_code.clone()),
+        };
+        match input.as_str().trim() {
+            "u8" => TypeInfo::UnsignedInteger(IntegerBits::Eight),
+            "u16" => TypeInfo::UnsignedInteger(IntegerBits::Sixteen),
+            "u32" => TypeInfo::UnsignedInteger(IntegerBits::ThirtyTwo),
+            "u64" => TypeInfo::UnsignedInteger(IntegerBits::SixtyFour),
+            "bool" => TypeInfo::Boolean,
+            "unit" => TypeInfo::Tuple(Vec::new()),
+            "byte" => TypeInfo::Byte,
+            "b256" => TypeInfo::B256,
+            "Self" | "self" => TypeInfo::SelfType,
+            "Contract" => TypeInfo::Contract,
+            _other => TypeInfo::Custom {
+                name: Ident::new(span),
+                type_arguments: vec![],
+            },
+        }
+    }
+
     pub(crate) fn friendly_type_str(&self) -> String {
         use TypeInfo::*;
         match self {
@@ -236,7 +442,7 @@ impl TypeInfo {
             }
             .into(),
             Boolean => "bool".into(),
-            Custom { name } => format!("unresolved {}", name.as_str()),
+            Custom { name, .. } => format!("unresolved {}", name.as_str()),
             Ref(id) => format!("T{} ({})", id, (*id).friendly_type_str()),
             Tuple(fields) => {
                 let field_strs = fields
@@ -254,17 +460,20 @@ impl TypeInfo {
             Enum {
                 name,
                 variant_types,
+                ..
             } => print_inner_types(
                 format!("enum {}", name),
                 variant_types.iter().map(|x| x.r#type),
             ),
-            Struct { name, fields } => {
-                print_inner_types(format!("struct {}", name), fields.iter().map(|x| x.r#type))
-            }
+            Struct { name, fields, .. } => print_inner_types(
+                format!("struct {}", name),
+                fields.iter().map(|field| field.r#type),
+            ),
             ContractCaller { abi_name, .. } => {
                 format!("contract caller {}", abi_name.suffix)
             }
             Array(elem_ty, count) => format!("[{}; {}]", elem_ty.friendly_type_str(), count),
+            Storage { .. } => "contract storage".into(),
         }
     }
 
@@ -282,7 +491,7 @@ impl TypeInfo {
             }
             .into(),
             Boolean => "bool".into(),
-            Custom { name } => format!("unresolved {}", name.as_str()),
+            Custom { name, .. } => format!("unresolved {}", name.as_str()),
             Ref(id) => format!("T{} ({})", id, (*id).json_abi_str()),
             Tuple(fields) => {
                 let field_strs = fields
@@ -307,6 +516,7 @@ impl TypeInfo {
                 format!("contract caller {}", abi_name.suffix)
             }
             Array(elem_ty, count) => format!("[{}; {}]", elem_ty.json_abi_str(), count),
+            Storage { .. } => "contract storage".into(),
         }
     }
 
@@ -332,7 +542,7 @@ impl TypeInfo {
                     let names = fields
                         .iter()
                         .map(|field_type| {
-                            resolve_type(*field_type, error_msg_span)
+                            resolve_type(field_type.type_id, error_msg_span)
                                 .expect("unreachable?")
                                 .to_selector_name(error_msg_span)
                         })
@@ -352,26 +562,22 @@ impl TypeInfo {
             Byte => "byte".into(),
             B256 => "b256".into(),
             Struct { fields, .. } => {
-                let field_names = {
-                    let names = fields
-                        .iter()
-                        .map(|TypedStructField { r#type, .. }| {
-                            resolve_type(*r#type, error_msg_span)
-                                .expect("unreachable?")
-                                .to_selector_name(error_msg_span)
-                        })
-                        .collect::<Vec<CompileResult<String>>>();
-                    let mut buf = vec![];
-                    for name in names {
-                        match name.value {
-                            Some(value) => buf.push(value),
-                            None => return name,
-                        }
+                let names = fields
+                    .iter()
+                    .map(|field| {
+                        resolve_type(field.r#type, error_msg_span)
+                            .expect("unreachable?")
+                            .to_selector_name(error_msg_span)
+                    })
+                    .collect::<Vec<CompileResult<String>>>();
+                let mut buf = vec![];
+                for name in names {
+                    match name.value {
+                        Some(value) => buf.push(value),
+                        None => return name,
                     }
-                    buf
-                };
-
-                format!("s({})", field_names.join(","))
+                }
+                format!("s({})", buf.join(","))
             }
             Enum { variant_types, .. } => {
                 let variant_names = {
@@ -408,9 +614,11 @@ impl TypeInfo {
         };
         ok(name, vec![], vec![])
     }
+
     pub(crate) fn size_in_bytes(&self, err_span: &Span) -> Result<u64, CompileError> {
         Ok(self.size_in_words(err_span)? * 8)
     }
+
     /// Calculates the stack size of this type, to be used when allocating stack memory for it.
     pub(crate) fn size_in_words(&self, err_span: &Span) -> Result<u64, CompileError> {
         match self {
@@ -423,7 +631,7 @@ impl TypeInfo {
             TypeInfo::Tuple(fields) => Ok(fields
                 .iter()
                 .map(|field_type| {
-                    resolve_type(*field_type, err_span)
+                    resolve_type(field_type.type_id, err_span)
                         .expect("should be unreachable?")
                         .size_in_words(err_span)
                 })
@@ -445,8 +653,8 @@ impl TypeInfo {
             }
             TypeInfo::Struct { fields, .. } => Ok(fields
                 .iter()
-                .map(|x| -> Result<_, _> {
-                    resolve_type(x.r#type, err_span)
+                .map(|field| -> Result<_, _> {
+                    resolve_type(field.r#type, err_span)
                         .expect("should be unreachable?")
                         .size_in_words(err_span)
                 })
@@ -469,6 +677,7 @@ impl TypeInfo {
             TypeInfo::Array(elem_ty, count) => {
                 Ok(look_up_type_id(*elem_ty).size_in_words(err_span)? * *count as u64)
             }
+            TypeInfo::Storage { .. } => Ok(0),
         }
     }
 
@@ -489,7 +698,7 @@ impl TypeInfo {
                 .any(|field| look_up_type_id(field.r#type).is_uninhabited()),
             TypeInfo::Tuple(fields) => fields
                 .iter()
-                .any(|field_type| look_up_type_id(*field_type).is_uninhabited()),
+                .any(|field_type| look_up_type_id(field_type.type_id).is_uninhabited()),
             _ => false,
         }
     }
@@ -527,7 +736,7 @@ impl TypeInfo {
             TypeInfo::Tuple(fields) => {
                 let mut all_zero_sized = true;
                 for field in fields {
-                    let field_type = look_up_type_id(*field);
+                    let field_type = look_up_type_id(field.type_id);
                     if field_type.is_uninhabited() {
                         return true;
                     }
@@ -556,7 +765,8 @@ impl TypeInfo {
         match self {
             TypeInfo::Custom { .. } => {
                 for (param, ty_id) in mapping.iter() {
-                    if param.name == *self {
+                    let param_type_info = look_up_type_id(param.type_id);
+                    if param_type_info == *self {
                         return Some(*ty_id);
                     }
                 }
@@ -564,8 +774,13 @@ impl TypeInfo {
             }
             TypeInfo::UnknownGeneric { name, .. } => {
                 for (param, ty_id) in mapping.iter() {
-                    if param.name == (TypeInfo::Custom { name: name.clone() }) {
-                        return Some(*ty_id);
+                    if let TypeInfo::Custom {
+                        name: other_name, ..
+                    } = look_up_type_id(param.type_id)
+                    {
+                        if name.clone() == other_name {
+                            return Some(*ty_id);
+                        }
                     }
                 }
                 None
@@ -596,7 +811,6 @@ impl TypeInfo {
                         new_variant.r#type = insert_type(TypeInfo::Ref(matching_id));
                     }
                 }
-
                 Some(insert_type(TypeInfo::Enum {
                     variant_types: new_variants,
                     name: name.clone(),
@@ -610,21 +824,28 @@ impl TypeInfo {
                 let mut index = 0;
                 while index < fields.len() {
                     let new_field_id_opt =
-                        look_up_type_id(fields[index]).matches_type_parameter(mapping);
+                        look_up_type_id(fields[index].type_id).matches_type_parameter(mapping);
                     if let Some(new_field_id) = new_field_id_opt {
                         new_fields.extend(fields[..index].iter().cloned());
-                        new_fields.push(insert_type(TypeInfo::Ref(new_field_id)));
+                        new_fields.push(TypeArgument {
+                            type_id: insert_type(TypeInfo::Ref(new_field_id)),
+                            span: fields[index].span.clone(),
+                        });
                         index += 1;
                         break;
                     }
                     index += 1;
                 }
                 while index < fields.len() {
-                    let new_field =
-                        match look_up_type_id(fields[index]).matches_type_parameter(mapping) {
-                            Some(new_field_id) => insert_type(TypeInfo::Ref(new_field_id)),
-                            None => fields[index],
-                        };
+                    let new_field = match look_up_type_id(fields[index].type_id)
+                        .matches_type_parameter(mapping)
+                    {
+                        Some(new_field_id) => TypeArgument {
+                            type_id: insert_type(TypeInfo::Ref(new_field_id)),
+                            span: fields[index].span.clone(),
+                        },
+                        None => fields[index].clone(),
+                    };
                     new_fields.push(new_field);
                     index += 1;
                 }
@@ -645,6 +866,7 @@ impl TypeInfo {
             | B256
             | Numeric
             | Contract
+            | Storage { .. }
             | ErrorRecovery => None,
         }
     }
